namespace Wilco.CodeHilighter.Engine.SyntaxHighlighting
{
    using System;
    using System.Drawing;
    using System.Linq;
    using System.Xml.Linq;

    /// <summary>
    /// Represents a comment line scanner.
    /// </summary>
    public class CommentLineScanner : ScannerBase
    {
        #region Fields

        private CommentLineNode commentLineNode;

        #endregion Fields

        #region Constructors

        /// <summary>
        /// Initializes a new instance of a <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.CommentLineScanner"/> class.
        /// </summary>
        public CommentLineScanner()
            : this(null, null, null)
        {
            //
        }

        /// <summary>
        /// Initializes a new instance of a <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.CommentLineScanner"/> class.
        /// </summary>
        /// <param name="tokenizer">The <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.TokenizerBase"/> which is used to tokenize the source code.</param>
        /// <param name="scannerResult">The <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.OccurrenceCollection"/> which will contain the scanner result.</param>
        public CommentLineScanner(TokenizerBase tokenizer, OccurrenceCollection scannerResult)
            : this(tokenizer, scannerResult, new CommentLineNode())
        {
            //
        }

        /// <summary>
        /// Initializes a new instance of a <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.CommentLineScanner"/> class.
        /// </summary>
        /// <param name="tokenizer">The <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.TokenizerBase"/> which is used to tokenize the source code.</param>
        /// <param name="scannerResult">The <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.OccurrenceCollection"/> which will contain the scanner result.</param>
        /// <param name="commentLineNode">A <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.CommentLineNode"/> object.</param>
        public CommentLineScanner(TokenizerBase tokenizer, OccurrenceCollection scannerResult, CommentLineNode commentLineNode)
            : base(tokenizer, scannerResult)
        {
            this.commentLineNode = commentLineNode;
            this.SetID("CommentLineScanner");
        }

        #endregion Constructors

        #region Properties

        /// <summary>
        /// Gets or sets the comment line node.
        /// </summary>
        public CommentLineNode CommentLineNode
        {
            get
            {
                return this.commentLineNode;
            }
            set
            {
                if (value != this.commentLineNode)
                {
                    this.commentLineNode = value;
                }
            }
        }

        #endregion Properties

        #region Methods

        /// <summary>
        /// Initializes a new instance of a <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.IScanner"/> implementation class.
        /// </summary>
        /// <param name="tokenizer">The <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.TokenizerBase"/> which is used to tokenize the source code.</param>
        /// <param name="scannerResult">The <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.OccurrenceCollection"/> which will contain the scanner result.</param>
        /// <returns>A new instance of a <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.IScanner"/> implementation class.</returns>
        public override IScanner Create(TokenizerBase tokenizer, OccurrenceCollection scannerResult)
        {
            return new CommentLineScanner(tokenizer, scannerResult);
        }

        /// <summary>
        /// Loads the state of the scanner.
        /// </summary>
        /// <param name="state">An <see cref="System.Object"/> that contains the state of the scanner.</param>
        public override void LoadState(object state)
        {
            XElement element = (XElement)state;

            // Load settings.
            this.commentLineNode.BackColor = ColorTranslator.FromHtml(Helper.GetAttributeValueFromElement(element, "settings", "setting","name", "BackColor"));
            this.commentLineNode.ForeColor = ColorTranslator.FromHtml(Helper.GetAttributeValueFromElement(element, "settings", "setting", "name", "ForeColor"));
            this.commentLineNode.Font = Helper.GetAttributeValueFromElement(element, "settings", "setting", "name", "Font");
            this.commentLineNode.NavigateUrl = Helper.GetAttributeValueFromElement(element, "settings", "setting", "name", "NavigateUrl");

            // Load entities.
            foreach (XElement entityElement in element.Descendants("entities").Descendants("entity"))
            {
                this.commentLineNode.Entities.Add(entityElement.Value);
            }
        }

        /// <summary>
        /// Saves the current state of the scanner.
        /// </summary>
        /// <param name="container">The container which will contain the state.</param>
        /// <returns>An <see cref="System.Object"/> that contains the state of the scanner.</returns>
        public override object SaveState(object container)
        {
            //XDocument document = (XDocument)container;
            XElement element = (XElement)base.SaveState(container);

            // Save settings.
            XElement settingRootElement = new XElement("settings");
            element.Add(settingRootElement);

            settingRootElement.Add(this.CreateSetting("BackColor", ColorTranslator.ToHtml(this.commentLineNode.BackColor)));
            settingRootElement.Add(this.CreateSetting("ForeColor", ColorTranslator.ToHtml(this.commentLineNode.ForeColor)));
            settingRootElement.Add(this.CreateSetting("Font", this.commentLineNode.Font));
            settingRootElement.Add(this.CreateSetting("NavigateUrl", this.commentLineNode.NavigateUrl));

            // Save entities.
            XElement entityRootElement = new XElement("entities");
            element.Add(entityRootElement);

            foreach (string entity in this.commentLineNode.Entities)
            {
                entityRootElement.Add(new XElement("entity", entity));
            }

            return element;
        }

        /// <summary>
        /// Scans a token.
        /// </summary>
        /// <remarks>
        /// An <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.IScanner"/> implementation will generally have a reference to a 
        /// <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.NodeCollection"/> which will be used to store results of a scan.
        /// </remarks>
        /// <param name="token">A token from the source code.</param>
        public override void Scan(string token)
        {
            if (!this.Enabled)
            {
                if (this.Child != null)
                {
                    this.Child.Scan(token);
                }
            }
            else
            {
                bool isMatch = false;

                for (int i = 0; i < this.commentLineNode.Entities.Count; i++)
                {
                    if ((this.Tokenizer.Position + this.commentLineNode.Entities[i].Length) <= this.Tokenizer.Source.Length)
                    {
                        if (this.Tokenizer.GetNextTokens(this.commentLineNode.Entities[i].Length) == this.commentLineNode.Entities[i])
                        {
                            isMatch = true;
                            int endOfLineIndex = this.Tokenizer.Source.IndexOf("\n", this.Tokenizer.Position);
                            if (endOfLineIndex == -1)
                                endOfLineIndex = this.Tokenizer.Source.Length;
                            else if (this.Tokenizer.Source[endOfLineIndex - 1] == '\r')
                                endOfLineIndex--;

                            this.ScannerResult.Add(new Occurrence(this.Tokenizer.Position, endOfLineIndex - this.Tokenizer.Position, this.commentLineNode));
                            this.Tokenizer.MoveTo(endOfLineIndex - 1);
                            break;
                        }
                    }
                }

                if (!isMatch)
                {
                    if (this.Child != null)
                    {
                        this.Child.Scan(token);
                    }
                }
            }
        }

        /// <summary>
        /// Creates a steting.
        /// </summary>
        /// <param name="document">The document which will contain the setting.</param>
        /// <param name="name">The name of the setting.</param>
        /// <param name="value">The value of the setting.</param>
        /// <returns>The <see cref="System.Xml.XmlElement"/> which represents the setting.</returns>
        private XElement CreateSetting(string name, string value)
        {
            return new XElement("setting", value, new XAttribute("name", name));
        }

        #endregion Methods
    }
}