namespace Wilco.CodeHilighter.Engine.SyntaxHighlighting
{
    using System;

    /// <summary>
    /// Represents a php scanner.
    /// </summary>
    public class PHPScanner : ScannerBase
    {
        #region Fields

        private bool inScriptBlock;
        private OccurrenceCollection scannerResult;
        private XmlScanner xmlScanner;

        #endregion Fields

        #region Constructors

        /// <summary>
        /// Initializes a new instance of a <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.PHPScanner"/> class.
        /// </summary>
        public PHPScanner()
            : base(null, null)
        {
            //
        }

        /// <summary>
        /// Initializes a new instance of a <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.PHPScanner"/> class.
        /// </summary>
        /// <param name="tokenizer">The <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.TokenizerBase"/> which is used to tokenize the source code.</param>
        /// <param name="scannerResult">The <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.OccurrenceCollection"/> which will contain the scanner result.</param>
        public PHPScanner(TokenizerBase tokenizer, OccurrenceCollection scannerResult)
            : base(tokenizer, scannerResult)
        {
            this.scannerResult = scannerResult;
            this.xmlScanner = new XmlScanner(this.Tokenizer, this.scannerResult);
            this.xmlScanner.Match += new MatchEventHandler(this.xmlScanner_Match);
            this.SetID("PHPScanner");
        }

        #endregion Constructors

        #region Methods

        /// <summary>
        /// Initializes a new instance of a <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.IScanner"/> implementation class.
        /// </summary>
        /// <param name="tokenizer">The <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.TokenizerBase"/> which is used to tokenize the source code.</param>
        /// <param name="scannerResult">The <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.OccurrenceCollection"/> which will contain the scanner result.</param>
        /// <returns>A new instance of a <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.IScanner"/> implementation class.</returns>
        public override IScanner Create(TokenizerBase tokenizer, OccurrenceCollection scannerResult)
        {
            return new PHPScanner(tokenizer, scannerResult);
        }

        /// <summary>
        /// Resets the scanner.
        /// </summary>
        public override void Reset()
        {
            this.inScriptBlock = false;
        }

        /// <summary>
        /// Scans a token.
        /// </summary>
        /// <remarks>
        /// An <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.IScanner"/> implementation will generally have a reference to a 
        /// <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.NodeCollection"/> which will be used to store results of a scan.
        /// </remarks>
        /// <param name="token">A token from the source code.</param>
        public override void Scan(string token)
        {
            if (!this.Enabled)
            {
                if (this.Child != null)
                {
                    this.Child.Scan(token);
                }
            }
            else
            {
                // Highlight as XML until a script block was found.
                if (!this.inScriptBlock)
                {
                    // Find script block.
                    this.xmlScanner.Scan(token);
                }
                else
                {
                    if ((this.Tokenizer.Position + 2) <= this.Tokenizer.Source.Length && this.Tokenizer.GetNextTokens(2) == "?>")
                    {
                        this.inScriptBlock = false;
                        this.xmlScanner.Scan(token);
                    }
                    else if (this.Child != null)
                        this.Child.Scan(token);
                }
            }
        }

        /// <summary>
        /// Handles the match event.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void xmlScanner_Match(object sender, MatchEventArgs e)
        {
            if (e.Type == MatchType.StartIdentifier)
            {
                this.inScriptBlock = true;
            }
            else if (e.Type == MatchType.EndIdentifier)
            {
                this.inScriptBlock = false;
            }
        }

        #endregion Methods
    }
}