namespace Wilco.CodeHilighter.Engine.SyntaxHighlighting
{
    using System;
    using System.Drawing;
    using System.Linq;
    using System.Xml.Linq;

    /// <summary>
    /// Represents a string block scanner.
    /// </summary>
    public class StringBlockScanner : ScannerBase
    {
        #region Fields

        private string lastToken;
        private StringNode stringNode;

        #endregion Fields

        #region Constructors

        /// <summary>
        /// Initializes a new instance of a <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.StringBlockScanner"/> class.
        /// </summary>
        public StringBlockScanner()
            : this(null, null, null)
        {
            //
        }

        /// <summary>
        /// Initializes a new instance of a <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.StringBlockScanner"/> class.
        /// </summary>
        /// <param name="tokenizer">The <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.TokenizerBase"/> which is used to tokenize the source code.</param>
        /// <param name="scannerResult">The <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.OccurrenceCollection"/> which will contain the scanner result.</param>
        public StringBlockScanner(TokenizerBase tokenizer, OccurrenceCollection scannerResult)
            : this(tokenizer, scannerResult, new StringNode())
        {
            //
        }

        /// <summary>
        /// Initializes a new instance of a <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.StringBlockScanner"/> class.
        /// </summary>
        /// <param name="tokenizer">The <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.TokenizerBase"/> which is used to tokenize the source code.</param>
        /// <param name="scannerResult">The <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.OccurrenceCollection"/> which will contain the scanner result.</param>
        /// <param name="stringNode">A <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.StringNode"/> object.</param>
        public StringBlockScanner(TokenizerBase tokenizer, OccurrenceCollection scannerResult, StringNode stringNode)
            : base(tokenizer, scannerResult)
        {
            this.stringNode = stringNode;
            this.lastToken = null;
            this.SetID("StringBlockScanner");
        }

        #endregion Constructors

        #region Properties

        /// <summary>
        /// Gets or sets the string node.
        /// </summary>
        public StringNode StringNode
        {
            get
            {
                return this.stringNode;
            }
            set
            {
                if (value != this.stringNode)
                {
                    this.stringNode = value;
                }
            }
        }

        #endregion Properties

        #region Methods

        /// <summary>
        /// Initializes a new instance of a <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.IScanner"/> implementation class.
        /// </summary>
        /// <param name="tokenizer">The <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.TokenizerBase"/> which is used to tokenize the source code.</param>
        /// <param name="scannerResult">The <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.OccurrenceCollection"/> which will contain the scanner result.</param>
        /// <returns>A new instance of a <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.IScanner"/> implementation class.</returns>
        public override IScanner Create(TokenizerBase tokenizer, OccurrenceCollection scannerResult)
        {
            return new StringBlockScanner(tokenizer, scannerResult);
        }

        /// <summary>
        /// Loads the state of the scanner.
        /// </summary>
        /// <param name="state">An <see cref="System.Object"/> that contains the state of the scanner.</param>
        public override void LoadState(object state)
        {
            XElement element = (XElement)state;

            // Load settings.
            this.stringNode.BackColor = ColorTranslator.FromHtml(Helper.GetAttributeValueFromElement(element, "settings", "setting","name", "BackColor"));
            this.stringNode.ForeColor = ColorTranslator.FromHtml(Helper.GetAttributeValueFromElement(element, "settings", "setting", "name", "ForeColor"));
            this.stringNode.Font = Helper.GetAttributeValueFromElement(element, "settings", "setting", "name", "Font");
            this.stringNode.NavigateUrl = Helper.GetAttributeValueFromElement(element, "settings", "setting", "name", "NavigateUrl");

            // Load entities.
            foreach (XElement entityElement in element.Descendants("entities").Descendants("entity"))
            {
                this.stringNode.Entities.Add(
                    new StringEntity(
                        entityElement.Attribute("start").Value,
                        entityElement.Attribute("end").Value,
                        entityElement.Attribute("escape").Value));
            }
        }

        /// <summary>
        /// Resets the scanner.
        /// </summary>
        public override void Reset()
        {
            this.lastToken = null;
        }

        /// <summary>
        /// Saves the current state of the scanner.
        /// </summary>
        /// <param name="container">The container which will contain the state.</param>
        /// <returns>An <see cref="System.Object"/> that contains the state of the scanner.</returns>
        public override object SaveState(object container)
        {
            //XDocument document = (XDocument)container;
            XElement element = (XElement)base.SaveState(container);

            // Save settings.
            XElement settingRootElement = new XElement("settings");
            element.Add(settingRootElement);

            settingRootElement.Add(this.CreateSetting("BackColor", ColorTranslator.ToHtml(this.stringNode.BackColor)));
            settingRootElement.Add(this.CreateSetting("ForeColor", ColorTranslator.ToHtml(this.stringNode.ForeColor)));
            settingRootElement.Add(this.CreateSetting("Font",this.stringNode.Font));
            settingRootElement.Add(this.CreateSetting("NavigateUrl", this.stringNode.NavigateUrl));

            // Save entities.
            XElement entityRootElement = new XElement("entities");
            element.Add(entityRootElement);
            XElement entityElement;
            foreach (StringEntity entity in this.stringNode.Entities)
            {
                entityElement = new XElement("entity");
                entityElement.SetAttributeValue("start", entity.Start);
                entityElement.SetAttributeValue("end", entity.End);
                entityElement.SetAttributeValue("escape", entity.Escape);
                entityRootElement.Add(entityElement);
            }

            return element;
        }

        /// <summary>
        /// Scans a token.
        /// </summary>
        /// <remarks>
        /// An <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.IScanner"/> implementation will generally have a reference to a 
        /// <see cref="Wilco.CodeHilighter.Engine.SyntaxHighlighting.NodeCollection"/> which will be used to store results of a scan.
        /// </remarks>
        /// <param name="token">A token from the source code.</param>
        public override void Scan(string token)
        {
            if (!this.Enabled)
            {
                if (this.Child != null)
                {
                    this.Child.Scan(token);
                }
            }
            else
            {
                bool isMatch = false;

                if (this.lastToken != null && this.lastToken == "@")
                {
                    for (int i = 0; i < this.stringNode.Entities.Count; i++)
                    {
                        if ((this.Tokenizer.Position + this.stringNode.Entities[i].Start.Length) <= this.Tokenizer.Source.Length)
                        {
                            if (this.Tokenizer.GetNextTokens(this.stringNode.Entities[i].Start.Length) == this.stringNode.Entities[i].Start)
                            {
                                this.Tokenizer.MoveTo(this.Tokenizer.Position + this.StringNode.Entities[i].Start.Length);
                                int startIndex = this.Tokenizer.Position;

                                int endOfLineIndex = this.Tokenizer.Source.Length;

                                string escape = ((StringEntity)this.StringNode.Entities[i]).Escape;

                                int index;
                                while (this.Tokenizer.Position <= endOfLineIndex)
                                {
                                    index = this.Tokenizer.Source.IndexOf(this.stringNode.Entities[i].End, this.Tokenizer.Position, endOfLineIndex - this.Tokenizer.Position);
                                    if (index > -1)
                                    {
                                        if (stringNode.Entities[i].End == escape && index + escape.Length < this.Tokenizer.Source.Length && this.Tokenizer.Source.Substring(index + escape.Length, escape.Length) == escape)
                                        {
                                            this.Tokenizer.MoveTo(index + escape.Length * 2);

                                            // Error.
                                            if (this.Tokenizer.Position == this.Tokenizer.Source.Length)
                                            {
                                                this.ScannerResult.Add(new Occurrence(startIndex, endOfLineIndex - startIndex, this.stringNode));
                                            }
                                        }
                                        else if (stringNode.Entities[i].End != escape && this.Tokenizer.Source.Substring(index - escape.Length, escape.Length) == escape)
                                        {
                                            this.Tokenizer.MoveTo(index + this.stringNode.Entities[i].End.Length);

                                            // Error.
                                            if (this.Tokenizer.Position == this.Tokenizer.Source.Length)
                                            {
                                                this.ScannerResult.Add(new Occurrence(startIndex, endOfLineIndex - startIndex, this.stringNode));
                                            }
                                        }
                                        else
                                        {
                                            // Success.
                                            this.ScannerResult.Add(new Occurrence(startIndex, index - startIndex, this.stringNode));
                                            this.Tokenizer.MoveTo(index + this.stringNode.Entities[i].End.Length);
                                            break;
                                        }
                                    }
                                    else
                                    {
                                        // Error.
                                        this.ScannerResult.Add(new Occurrence(startIndex, endOfLineIndex - startIndex, this.stringNode));
                                        this.Tokenizer.MoveTo(endOfLineIndex - 1);
                                        break;
                                    }
                                }

                                isMatch = true;
                                break;
                            }
                        }
                    }
                }

                this.lastToken = token;

                if (!isMatch)
                {
                    if (this.Child != null)
                    {
                        this.Child.Scan(token);
                    }
                }
            }
        }

        /// <summary>
        /// Creates a steting.
        /// </summary>
        /// <param name="document">The document which will contain the setting.</param>
        /// <param name="name">The name of the setting.</param>
        /// <param name="value">The value of the setting.</param>
        /// <returns>The <see cref="System.Xml.XmlElement"/> which represents the setting.</returns>
        private XElement CreateSetting(string name, string value)
        {
            return new XElement("setting",value, new XAttribute("name", name));
        }

        #endregion Methods
    }
}